from ._air_scattering_coefficient import (
    AirScatteringCoefficientSpectrum as AirScatteringCoefficientSpectrum,
)
from ._core import Spectrum as Spectrum
from ._core import spectrum_factory as spectrum_factory
from ._interpolated import InterpolatedSpectrum as InterpolatedSpectrum
from ._multi_delta import MultiDeltaSpectrum as MultiDeltaSpectrum
from ._solar_irradiance import SolarIrradianceSpectrum as SolarIrradianceSpectrum
from ._uniform import UniformSpectrum as UniformSpectrum
